/*
 * Copyright © 2009 Jerome Glisse <glisse@freedesktop.org>
 *
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include "xf86drm.h"
#include "xf86drmMode.h"
#include "radeon_drm.h"
#include "radeon.h"
#include "radeon_bo.h"
#include "radeon_mode.h"

const char *connector_status(drmModeConnector *connector)
{
    switch (connector->connection) {
    case DRM_MODE_CONNECTED:
        return "connected";
    case DRM_MODE_DISCONNECTED:
    default:
        return "disconnected";
    }
    return "unknown";
}

int radeon_mode_configure(int radeonfd, struct radeon_mode *mode)
{
    drmModeRes *resources;
    drmModeConnector *connector;
    drmModeEncoder *encoder;
    uint32_t connector_id;
    unsigned i, j;
    int r;

    r = drmSetMaster(radeonfd);
    if (r) {
        fprintf(stderr, "Failed to set master\n");
        perror(NULL);
    }
    mode->bo = NULL;
    mode->fb_id = 0;
    resources = drmModeGetResources(radeonfd);
    if (resources == NULL) {
            fprintf(stderr, "Failed to get resources from card\n");
            return -EINVAL;
    }
    /* Iterates through connector an pick first one with a valid mode */
    printf("%d connectors\n", resources->count_connectors);
    for (i = 0; i < resources->count_connectors; i++) {
        connector = drmModeGetConnector(radeonfd, resources->connectors[i]);
        if (connector == NULL) {
            fprintf(stderr, "Could not get connector %i\n",
                    resources->connectors[i]);
            r = -EINVAL;
            goto out;
        }
        printf("connector(%d) %s to encoder(%d)\n", connector->connector_id,
               connector_status(connector), connector->encoder_id);
        for (j = 0; j < connector->count_modes; j++) {
            if (connector->modes[j].type & DRM_MODE_TYPE_PREFERRED) {
                memcpy(&mode->info, &connector->modes[j],
                       sizeof(drmModeModeInfo));
                mode->connector_id = connector->connector_id;
                mode->encoder_id = connector->encoder_id;
                drmModeFreeConnector(connector);
                goto setmode;
            }
        }
        drmModeFreeConnector(connector);
    }
    fprintf(stderr, "No modes found !\n");
    r = -EINVAL;
    goto out;
setmode:
    /* Find crtc id */
    encoder = drmModeGetEncoder(radeonfd, mode->encoder_id);
    if (encoder == NULL) {
        fprintf(stderr, "Could not get encoder %i\n", mode->encoder_id);
        r = -EINVAL;
        goto out;
    }
    mode->crtc_id = resources->crtcs[ffs(encoder->possible_crtcs)-1];
    drmModeFreeEncoder(encoder);
    /* Allocate buffer */
    mode->bpp = 4;
    mode->npix_y = mode->info.vdisplay;
    mode->npix_x = mode->info.hdisplay;
    mode->pitch = mode->info.hdisplay;
    mode->pitch = (mode->pitch + 255) & ~255;
    mode->pitch *= mode->bpp;
    mode->size = mode->pitch * mode->info.vdisplay;
    mode->bo = radeon_bo(radeonfd, 0, mode->size, 4096, NULL);
    if (mode->bo == NULL) {
        fprintf(stderr, "failed to create fb: %s\n", strerror(errno));
        goto out;
    }
    r = radeon_bo_map(mode->bo);
    if (r) {
        fprintf(stderr, "failed to map fb: %s\n", strerror(errno));
        goto out;
    }
    memset(mode->bo->data, 0x66, mode->size);
    r = drmModeAddFB(radeonfd, mode->info.hdisplay, mode->info.vdisplay,
                     32, 32, mode->pitch, mode->bo->handle, &mode->fb_id);
    if (r) {
        fprintf(stderr, "failed to add fb: %s  0x%08X\n", strerror(errno), mode->bo->handle);
        goto out;
    }
    printf("Mode: %dx%d %dbpp on CRTC(%d) connector(%d)\n", mode->info.hdisplay,
           mode->info.vdisplay, mode->bpp*8, mode->crtc_id,
           mode->connector_id);
    connector_id = mode->connector_id;
    r = drmModeSetCrtc(radeonfd, mode->crtc_id, mode->fb_id, 0, 0,
                       &connector_id, 1, &mode->info);
    if (r) {
        fprintf(stderr, "failed to set mode: %s\n", strerror(errno));
        goto out;
    }
    if (connector_id != mode->connector_id) {
        fprintf(stderr, "Set mode on different connector %d, expected %d\n",
                connector_id, mode->connector_id);
    }
out:
    if (r && mode->fb_id) {
        drmModeRmFB(radeonfd, mode->fb_id);
        mode->fb_id = 0;
    }
    if (r && mode->bo) {
        mode->bo = radeon_bo_decref(mode->bo);
    }
    drmModeFreeResources(resources);
    return r;
}

void radeon_mode_cleanup(int radeonfd, struct radeon_mode *mode)
{
    if (mode->fb_id) {
        drmModeRmFB(radeonfd, mode->fb_id);
        mode->fb_id = 0;
    }
    if (mode->bo) {
        radeon_bo_unmap(mode->bo);
        mode->bo = radeon_bo_decref(mode->bo);
    }
    drmDropMaster(radeonfd);
}
